! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
! catcorr_aux.f90 Original author: Clive Page, cgp@leicester.ac.uk
! Revised CGP 2012 June 14: v0.4.4: improved error messages & documentation
! Revised CGP 2012 Aug 23:  v0.4.5: fixed NMATCHES output
! Revised CGP 2012 Aug 23:  v0.4.6: fixed HISTORY keyword time value.
! Revised CGP 2012 Aug 29:  v0.4.7: added REFCAT keyword, DIAG parameter
! Revised CGP 2012 Oct  4:  v0.4.8: additional diagnostic output
! Revised CGP 2012 Oct 11:  v0.4.8: can force catalogue via environment var
! Revised CGP 2012 Oct 25:  v0.4.8: changed SDSS version to DR9
! Revised CGP 2012 Nov  5:  v0.5: added task params syserr, maxpofferr
! Revised CGP 2012 Nov 23:  v0.5.1: reports catalogue data release in OBSMLI
! Revised CGP 2012 Nov 30:  v0.6:   SYSERRCC column added 
! Revised CGP 2012 Dec 19:  v0.7:   optionally apply astrometric corrections
module catcorr_aux
use types
use errorHandling
use constants	       ! in ssclib, for Pi and TwoPi
use catcorr_minor	! defined in catprep
use nelder_mead
use sla              ! defined in sla
implicit none
private
interface get_fits_keyword
   module procedure get_fits_keyword_c, get_fits_keyword_d, &
    get_fits_keyword_e, get_fits_keyword_j, get_fits_keyword_l
end interface
public catcorr_top
character(*), parameter :: VERSION = "0.7"
logical, parameter      :: DIAG    = .false.  ! Produces diagnosics file if true
integer, parameter      :: LOG     = 9        ! I/O unit for diagnostic output
real(double), parameter :: HALFPI = 0.5_double * PI
real(double), parameter :: DTOR = PI/180.0_double, RTOD = 180.0_double/PI
real(double), parameter :: ASTOR = pi/(180.0_double * 3600.0_double), &
                           RTOAS = 1.0_double/ASTOR
integer, parameter  :: MAXSRC =   5000  ! Max number of X-ray detections
integer, parameter  :: MAXOPT = 100000  ! Max objects in any optical (or IR) catalogue
integer, parameter  :: MAXCAT =      3  ! Number of reference catalogues
integer, parameter  :: MAXCP  =    100  ! Max opt/IR counterparts per X-ray detn
real(double), parameter :: CATRADIUS = 14.0 * 60.0 * ASTOR  ! radius of field of view (rads)
real(double), parameter :: COINC_RADIUS = 60.0 * ASTOR      ! radius for count of object density

type XlistType       ! structure for X-ray detections from one observation
                     ! derived from XMM FITS file of type OBSMLI
   character(10):: obs_id = ' '    ! observation identifier
   real(double) :: ra_pnt  = 0.0_double       ! RA of pointing axis (radians)
   real(double) :: dec_pnt = 0.0_double       ! Dec of Pointing axis (radians)
   real         :: year    = 2000.0           ! Years of obsvn e.g. 2009.3
   integer      :: nsrc = 0                   ! Number of sources in following arrays
   integer      :: badflag(MAXSRC) = 0           ! nonzero for poor detections
   integer      :: src_num(MAXSRC) = 0           ! Source number within this file
   real(double) :: ra(MAXSRC) = 0.0_double       ! RA from emldetect, J2000 (radians)
   real(double) :: dec(MAXSRC) = 0.0_double      ! Dec ditto , J2000 (radians)
   real(double) :: lii(MAXSRC) = 0.0_double      ! Galactic longitude (radians)
   real(double) :: bii(MAXSRC) = 0.0_double      ! Galactic latitude (radians)
   real         :: radec_err(MAXSRC) = 0.0       ! Standard error in position (radians)
   real         :: ep_extent(MAXSRC) = 0.0       ! Extent (pixels) =0.0 for point sources
   real         :: ep_tot(MAXSRC) = 0.0          ! EP_TOT total counts/sec
   real(double) :: offax(MAXSRC) = 0.0           ! mean off-axis angle (radians)
! Following items set from scan of reference catalogues for potential cross-matches
   integer      :: ncp(MAXSRC,MAXCAT)            ! No of counterparts in each reference catalogue
   integer      :: listcp(MAXCP,MAXSRC,MAXCAT)   ! List of counterparts in ditto.
   integer      :: density(MAXSRC,MAXCAT)        ! No of objects within 1 arcmin
! Following items record catcorr results for each catalogue
   integer      :: nmatches(MAXCAT) = 0          ! No of optical/ir matches used
   real         :: goodzero(MAXCAT) = 0.0        ! goodness at zero frame shift
   real         :: goodness(MAXCAT) = 0.0        ! Goodness of fit or likelihood
   real(double) :: offset(3,MAXCAT) = 0.0        ! RA/dec offsets, ROT_COR (RADIANS)
   real         :: offerr(3,MAXCAT) = 0.0        ! errors in above (RADIANS)
   real         :: netprob(MAXSRC,MAXCAT) = 0.0  ! Net probability
! Following items record results for chosen best reference catalogue
   logical      :: poscorok = .false.            ! .true. if catcorr successful
   real(double) :: ra_cc(MAXSRC) = 0.0_double    ! Corrected ra
   real(double) :: dec_cc(MAXSRC) = 0.0_double   ! Corrected dec
   real(double) :: lii_cc(MAXSRC) = 0.0_double   ! Galactic longitude ditto (radians)
   real(double) :: bii_cc(MAXSRC) = 0.0_double   ! Galactic latitude ditto (radians)
   real         :: syserrcc(MAXSRC) = 0.0        ! Detn posn error (arcsec) est by CATCORR
   character(20):: catused = '        '          ! catalogue used or failure reason
end type XlistType
type(XlistType), save :: xlist
!
type catType          ! Storage for objects in the reference catalogues
   integer      :: nobj = 0             ! Number of catalogue objects in these arrays:
   character(10):: extname              ! Extension name
   real(double) :: cat_ra(MAXOPT)  = 0.0_double ! RA, J2000 (radians) 
   real(double) :: cat_dec(MAXOPT) = 0.0_double ! Dec, J2000 (radians)
   real         :: perr(MAXOPT) = 0.0   ! Std error on source postion (radians)
end type catType
!
type(catType), save :: CATS(MAXCAT) ! storage for up to 3 catalogue extract,
! typically but not necessarily (1) USNO, (2) 2MASS, (3) SDSS
!
integer  :: NCAT = 0            ! Current catalogue used for fiting with MINIM
integer  :: NMATCHES = 0        ! No of possible X-ray-opt matches
real     :: SYSERRSQ            ! Square of systematic error [radians^2]
!
CONTAINS
!--------------------------------------------------------------------------
! apply_astcorr: computed modified (ra,dec) with astrometry correction
subroutine apply_astcorr
integer :: irow
real(double) :: dec_ref, vpnt(3), vref(3), ctos(3,3), stoc(3,3), vx(3), &
                newvx(3), factor, newcel(3), newra, newdec
logical, parameter :: astcorr_diag = .false.
integer, parameter :: alog = 8
!
if(astcorr_diag) then
   open(unit=alog, file='P'//xlist%obs_id//'.log', status='replace')
end if
!
! need to find a reference point perpendicular to the pointing axis, 
! might as well use same RA just adjust the DEC
!
if(xlist%dec_pnt > 0.0) then
   dec_ref = xlist%dec_pnt - 0.5_double * pi
else
   dec_ref = xlist%dec_pnt + 0.5_double * pi
end if
call cona2v(xlist%ra_pnt, xlist%dec_pnt, vpnt)
call cona2v(xlist%ra_pnt, dec_ref,       vref)
!
! Use these two 3-vectors to compute coordinate rotation matrix and inverse
!
call congen(vpnt, vref, ctos)
call conmin(ctos, stoc)
if(astcorr_diag) then
   write(alog, '(a,2f12.6)') 'RA_PNT, DEC_PNT=', xlist%ra_pnt*rtod, xlist%dec_pnt*rtod
   write(alog, '(2a)')' row   RA        DEC     offaxis      vx(1)     vx(2)     ', &
   'vx(3)     factor   newvx(1)  newvx(2)  newvx(3)    newra      newdec'
end if
do irow = 1, xlist%nsrc
   call cona2v(xlist%ra(irow), xlist%dec(irow), vx)
   call rotvec(vx, ctos, newvx)
   factor = astcorr_factor(xlist%offax(irow))
!
! Apply this factor to the x, and y elements of newvx(1:3), then normalise
!
   newvx(1:2) = newvx(1:2) * factor
   newvx(3) = sqrt(1.0_double - newvx(1)**2 - newvx(2)**2)
!
! Rotate back to the celestial frame
!
   call rotvec(newvx, stoc, newcel)
   call conv2a(newcel, newra, newdec)
   if(astcorr_diag) then
       write(alog, '(i4,2f11.6, 8f10.6, 2f11.6)') &
          irow, xlist%ra(irow)*rtod, xlist%dec(irow)*rtod, &
          xlist%offax(irow)*rtod*60.0, &
          vx, factor, newvx, newra*rtod, newdec*rtod
   end if
   xlist%ra(irow)  = newra
   xlist%dec(irow) = newdec
! Also update the corresponding galactic coordinates
   call sla_eqgal(newra, newdec, xlist%lii(irow), xlist%bii(irow))

end do

! Ensure that the corrected coords also apply astcorr in case the poscorr
! correction fails (SPR-7697)
xlist%ra_cc(1:xlist%nsrc)  = xlist%ra(1:xlist%nsrc)
xlist%dec_cc(1:xlist%nsrc) = xlist%dec(1:xlist%nsrc)
xlist%lii_cc(1:xlist%nsrc) = xlist%lii(1:xlist%nsrc)
xlist%bii_cc(1:xlist%nsrc) = xlist%bii(1:xlist%nsrc)

if(astcorr_diag) close(unit=alog)

end subroutine apply_astcorr
!--------------------------------------------------------------------------
! astcorr_factor: returns multiplier of off-axis angle for astrometry correction
real(double) function astcorr_factor(offaxis)
real(double), intent(in) :: offaxis   ! Off-axis angle [RADIANS]
!
! The following coefficients are from a cubit fit by Andy Read, 2012 Dec 14
!
real :: c(0:3) = (/+0.0158736, -0.036454, +0.00942358, -0.000277756 /)
real :: r, yoff, yoffrad
!
! Convert off-axis angle from radians to arcminutes
!
r = offaxis * RTOD * 60.0_double
yoff = c(0) + c(1) * r + c(2) * r**2 + c(3) * r**3
!
! Note: the yoff value is a positive number of arcseconds, this needs to be
! subtracted from the observed offset (which is in radians) 
! and converted to a fractional change in off-axis angle
!
yoffrad = dtor * (yoff / 3600.0_double)
astcorr_factor = (offaxis - yoffrad) / offaxis
end function astcorr_factor
!--------------------------------------------------------------------------
subroutine catcorr_top(infile, catextract, mingood, minfit, maxOffset, &
                       syserr, maxpofferr, astcorr, status)
! Finds bets frame fit for one OBSMLI file using 3 catalogues
! If catextract FITS file is provided it uses that, else searches 3 cats itself
character(*), intent(in) :: infile     ! OBSMLI FITS file
character(*), intent(in) :: catextract ! Catalogue extract file (or blank)
integer, intent(in)      :: mingood    ! Minimum number of good detections
real, intent(in)         :: minfit     ! Minimum goodness of fit value to accept
real, intent(in)         :: maxOffset  ! Max frame offset to accept [arcsec]
real, intent(in)         :: syserr     ! Notional systematic positional error
                                       ! between X-ray and ref cat [arcsec
real, intent(in)         :: maxpofferr ! Maximum position offset error [arcsec]                                       
logical, intent(in)      :: astcorr    ! whether astrometric corrections to be done
integer, intent(inout)   :: status     ! Inherited status
!
integer, parameter  :: nop = 3, maxfn = 2000, iprint = -1, nloop = 6, iquad = 1
real(double), parameter :: stopcr = 1.0e-2_double, simp = 1.0e-5_double
real(double), parameter :: radius = 17.0_double * 60.0_double * ASTOR ! FOV radius, arcsecs
character(10) :: cdate, ctime, usecat
character(100) :: diagfile
integer       :: kslash, nxgood, ifault, nxsrc, isrc, forcedcat, bestcat, &
                 catmatches(MAXCAT)
real(double)  :: bestfit, var(nop), param(nop), step(nop), off, offerr, cosdec, &
             fit0(MAXCAT)
!
if(status /= 0) RETURN
SYSERRSQ = (syserr*ASTOR)**2  ! square of systematic error [radians]

forcedcat = 0
! if DIAG set: open file for diagnostic output using path of OBSMLI file if any
if(DIAG) then
   call get_environment_variable('catcorr_usecat', usecat)
   kslash = index(infile, '/', .true.)
   diagfile = infile(1:kslash) // 'catcorr_diag_u' // trim(usecat) // '.txt'
   open(unit=LOG, file=diagfile, status='replace')
   call date_and_time(cdate, ctime)
   write(log,*)'catcorr version ', version, ' run ', cdate, ctime
   write(*,*)'catcorr version ', version, ' run ', cdate, ctime
   write(log,*) 'mingood=', mingood, 'minfit=', minfit, &
      'maxOffset=', maxoffset, 'syserr=', syserr, &
      'maxPoffErr=', maxpofferr 
end if
!
! Set up default values for cases where the fit fails
!
xlist%offset     = 0.0_double
xlist%offerr     = 0.0_double
xlist%goodness   = 0.0_double
catmatches = 0
!
! Read list of X-ray detections from OBSMLI data product file
!
call OBSMLI_read(infile, xlist, status)
if(astcorr) call apply_astcorr
!
! Check if any lines of detections exist, if so exclude them from use
!
call hist_angle(xlist, status)
!
nxsrc  = xlist%nsrc
nxgood = count(xlist%badflag(1:nxsrc) == 0)
if(nxgood < mingood) then
   xlist%poscorok = .false.
   xlist%catused = 'too few matches'
   NCAT    = 1   ! arbitrary choice so zeros get written to OBSMLI file
   bestcat = 1   ! ditto
else
!
! read optical/IR data from up to 3 catalogues
!
   call cat_read_extracts(catextract, CATS, status)
!
! To speed up cross-match, find for each xray source all catalogue sources
! nearby and store the information in module storage for cross_match
!
   call cat_find_all_pairs(xlist, maxOffset, CATS)
!
! For each catalogue with enough objects, match X-ray list and fit frame shift
!
   do NCAT = 1, MAXCAT
      if(CATS(NCAT)%nobj < 100) CYCLE
!
! start with offsets zero so results not biased by eposcorr fitting
!
      param = 0.0
      call cross_match(param, fit0(NCAT))
      fit0(NCAT) = abs(fit0(NCAT))
      xlist%goodzero(NCAT) = fit0(NCAT)
      step   = (/0.1_double*ASTOR, 0.1_double*ASTOR, 0.05_double*dtor/)  
      call minim(param, step, nop, bestfit, maxfn, iprint, stopcr, nloop, &
              iquad, simp, var, cross_match, ifault)
      if(ifault == 0) then
         xlist%goodness(NCAT)   = abs(bestfit)
         xlist%offset(:,NCAT)   = param
         xlist%offerr(:,NCAT)   = sqrt(var)
         xlist%nmatches(NCAT) = NMATCHES
      end if
   end do
!
! Choose the catalogue which produced best goodness of fit
!
   bestcat = maxval(maxloc(xlist%goodness))  ! uses maxval() because maxloc returns array
   NCAT = bestcat
! In diagnostic mode check environement variable to see if want to force cat choice
   if(diag) then
      write(log,*)    'Catalogue with best fit ', bestcat, cats(bestcat)%extname
      if(usecat /= ' ') then
         read(usecat, *) forcedcat
         if(forcedcat > 0) then
! Need to check that this catalogue had some results
            if(xlist%goodness(forcedcat) > 0.1) then
               bestcat = forcedcat
               write(log,*) 'Forced use of catalogue ', bestcat, cats(bestcat)%extname
            end if
         end if
      end if
   end if      
   off  = sqrt((xlist%offset(1,bestcat)*cosdec)**2 + &
                xlist%offset(2,bestcat)**2) * RTOAS
   offerr = sqrt((xlist%offerr(1,bestcat)*cosdec)**2 + &
                xlist%offerr(2,bestcat)**2) * RTOAS
   if(xlist%goodness(bestcat) < minfit) then
      xlist%offerr(:,bestcat) = 0.0_double
      xlist%poscorok = .false.
      xlist%catused = 'poor fit'
   else if(off > maxOffset) then
      xlist%offerr(:,bestcat) = 0.0_double
      xlist%poscorok = .false.      
      xlist%catused = 'excessive offset'
   else if(offerr > maxpofferr) then
      xlist%offerr(:,bestcat) = 0.0_double
      xlist%poscorok = .false.
      xlist%catused = 'excessive perror'
   else                ! Everything ok, store results
      cosdec    = cos(xlist%dec_pnt)
      xlist%poscorok = .true.
      xlist%catused = cats(bestcat)%extname
!
! save revised positions in data structure
!
      do isrc = 1,xlist%nsrc
         call rot_shift(xlist%ra_pnt, xlist%dec_pnt, xlist%offset(:,bestcat), &
            xlist%ra(isrc), xlist%dec(isrc), &
            xlist%ra_cc(isrc), xlist%dec_cc(isrc))
! And calculate galactic coordinates of these
         call sla_eqgal(xlist%ra_cc(isrc), xlist%dec_cc(isrc), &
                        xlist%lii_cc(isrc), xlist%bii_cc(isrc))
      end do  ! isrc
   end if
   if(forcedcat > 0) NCAT = forcedcat
end if
!
! Estimate the position error for each detection, note %offerr values all radians
!
if(xlist%poscorok) then
   do isrc = 1, xlist%nsrc
      xlist%syserrcc(isrc) = &
        sqrt( (cos(xlist%dec_pnt) * xlist%offerr(1,bestcat))**2 +  &
                                    xlist%offerr(2,bestcat)**2  +  &
               (xlist%offax(isrc) * xlist%offerr(3,bestcat))**2 ) * rtoas
   end do
else
   !xlist%syserrcc(1:xlist%nsrc) = 1.5    ! default where no position correction possible
   xlist%syserrcc(1:xlist%nsrc) = 1.29    ! default where no position correction possible
end if                  
!
! update the data structure and OBSMLI file 
!
!call print_summary
if(DIAG) call print_summary ! RDS
if(DIAG) call print_detlist(bestcat)

call OBSMLI_update(infile, bestcat, astcorr, status)
if(diag) write(*,*) 'Diagnostic output written to: ', diagfile
end subroutine catcorr_top
!-----------------------------------------------------------------------------
! cat_find_all_pairs: finds list of potential matches between xray and opt/ir sources
subroutine cat_find_all_pairs(xlist, maxOffset, cats)
type(xlistType), intent(inout) :: xlist     ! list of X-ray detections
real, intent(in)               :: maxOffset ! Max offset to allow [arcsec]
type(catType), intent(IN)      :: cats(:)   ! Data from optical/ir catalogues
!
integer      :: icat, ix, iobj, ncoinc, n, npairs(MAXCAT)
real(double) :: xra, xdec, offset, maxPairOffset
!
! Zero the output arrays which will accumulate match references
!
xlist%ncp    = 0
xlist%listcp = 0
npairs       = 0  ! counter of counterparts for message at end
! Store pairs with offsets up to maxOffset + 5 arcsecs, to cover the
! largest potential error circle.  Work through all "good" X-ray detections.
maxPairOffset = (maxOffset + 5.0) * ASTOR
do icat = 1, MAXCAT
   do ix = 1, xlist%nsrc
      if(xlist%badflag(ix) > 0) CYCLE
      ncoinc = 0
      xra = xlist%ra(ix)
      xdec = xlist%dec(ix)
      do iobj = 1, cats(icat)%nobj
         offset = gcdist(xra, xdec, cats(icat)%cat_ra(iobj), &
                                    cats(icat)%cat_dec(iobj))
! Count field objects within about 1 arcminute of the x-ray position
         if(offset <= COINC_RADIUS) then
            ncoinc = ncoinc + 1
         end if
! Now see if offset is small enough for this to be a potential counterpart
      	if(offset <= maxPairOffset) then
            if(xlist%ncp(ix,icat) >= MAXCP) then
               call warning('tooManyPairs', 'cat_find_all_pairs - too many pairs')
   	         EXIT
   	      end if
   	      n = xlist%ncp(ix,icat) + 1
            xlist%listcp(n, ix, icat) = iobj
            xlist%ncp(ix,icat)        = n
            npairs(icat) = npairs(icat) + 1
         end if
      end do ! iobj
      xlist%density(ix,icat) = max(ncoinc - 1, 0)
   end do    ! ix
   if(diag) write(log,'(a,i6, a,i5,a)') cats(icat)%extname, cats(icat)%nobj, &
                  ' objects', npairs(icat), ' potential matches'
end do       ! icat
end subroutine cat_find_all_pairs
!-----------------------------------------------------------------------------
subroutine cat_read_extracts(filename, cat, status)
character(*), intent(in)   :: filename ! FITS file containing catalogue extracts
type(catType), intent(out) :: cat(:)   ! Catalogue data structure (up to 3 sets)
integer, intent(inout)     :: status   ! Inherited status
!
integer :: unit, icat, hdutype, nrows, kra, kdec, kperr 
logical :: anyf
!
if(status /= 0) RETURN
! Open the file READONLY
call ftgiou(unit, status)
call ftnopn(unit, filename, 0, status)
if(status /= 0) call error('missingExtract', &
            'cannot open catalogue extracts file '//trim(filename))
do icat = 1,3
! Move to HDU number (icat+1) containing a binary table, get number of rows
   call ftmahd(unit, icat+1, hdutype, status)
   if(status /= 0) then
! This HDU does not exist, set no of rows for the rest to zero, exit
      status = 0
      cat(icat:)%nobj = 0
      cat(icat:)%extname = ' '
      EXIT
   end if
   call ftgnrw(unit, nrows, status)
! Get the EXTNAME value of this HDU
   call get_fits_keyword(unit, 'EXTNAME', cat(icat)%extname, status)
! Read in the arrays, assumed to be in the order specified by cat_write_extracts
   cat(icat)%nobj = nrows
   if(nrows > 0) then
      call get_fits_column(unit, 'CAT_RA',        kra,  status)
      call get_fits_column(unit, 'CAT_DEC',       kdec, status)
      call get_fits_column(unit, 'CAT_RADEC_ERR', kperr, status)
      call ftgcvd(unit, kra,   1, 1, nrows, 0.0d0, cat(icat)%cat_ra(1:nrows),   anyf, status)
      call ftgcvd(unit, kdec,  1, 1, nrows, 0.0d0, cat(icat)%cat_dec(1:nrows),  anyf, status)
      call ftgcve(unit, kperr, 1, 1, nrows, 0.0,   cat(icat)%perr(1:nrows), anyf, status)
! Convert ra, dec, position error from degrees to radians
      cat(icat)%cat_ra(1:nrows)  = cat(icat)%cat_ra(1:nrows)  * DTOR
      cat(icat)%cat_dec(1:nrows) = cat(icat)%cat_dec(1:nrows) * DTOR
      cat(icat)%perr(1:nrows)    = cat(icat)%perr(1:nrows)    * DTOR
   end if
end do
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('cat_read_extracts', status)
end subroutine cat_read_extracts
!------------------------------------------------------------------------------
!+circle_around finds coordinates of NPTS points around a given point
subroutine circle_points(ra, dec, radius, rapt, decpt)
real(double), intent(in)  :: ra, dec, radius   ! (ra,dec) of centre, radius [radians]
real(double), intent(out) :: rapt(:), decpt(:) ! ra,dec of points (each size NPTS)
! in circle [radians] note wrapped so rapt() is close to ra even if -ve or
! over 2*pi radians in size, for convenience in finding map to use.
!
real(double) :: vz(3), vx(3), ctos(3,3), stoc(3,3), az, el, rapoint, decpoint
integer :: j, npts
!
! Get matrix to a new frame with this position at its centre, meridian arbitrary
!
npts = size(rapt)
call cona2v(ra, dec, vz)
call cona2v(ra, dec-0.5_double, vx)
call congen(vz, vx, ctos)
call conmin(ctos, stoc)
! Generate N points around a circle offset by radius from the pole
do j = 1,npts
   az = (twopi/npts) * j
   el = halfpi - radius
   call convrt(az, el, stoc, rapoint, decpoint)
! Wrap the RA value so that it is near the centre RA even if -ve or over 2pi
   rapt(j)  = mod(rapoint-ra + pi + twopi, twopi) + ra - pi
   decpt(j) = decpoint
end do
end subroutine circle_points
!-----------------------------------------------------------------------------
! cross_match: finds goodness of fit between X-ray detections and optical/IR
! counterparts, for current frame offset and reference catalogue NCAT
! Note: this routine is called for each iteration of MINIM
subroutine cross_match(param, result)
real(double), intent(in) :: param(:)  ! offsets in ra, dec, rotation (RADIANS)
real(double), intent(out) :: result   ! evaluated function at these offsets
!
real(double), parameter :: max_fiterr = 2.0_double ! arcsecs
integer :: ix, icp, icat
real(double) :: pmatch, pcoinc, net_prob, goodness, best_prob, ra2, dec2
!
goodness = 0.0_double  ! Accumulates sum of individual net probabilities
NMATCHES = 0       ! Counts no of possible matches for export to FITS file
! Loop over all X-ray detections in the list
! But skip poor quality detections and those with zero potential counterparts
do ix = 1, xlist%nsrc
   if(xlist%ncp(ix,NCAT) == 0 .or. xlist%badflag(ix) > 0) CYCLE
!
! Apply the current parameter offset and rotation to the coordinates of this source
!
   call rot_shift(xlist%ra_pnt, xlist%dec_pnt, param, xlist%ra(ix), &
                  xlist%dec(ix), ra2, dec2)
! Loop over all potential opt/IR counterparts for this X-ray detection
   best_prob = 0.0_double
   do icp = 1, xlist%ncp(ix,NCAT)
      icat = xlist%listcp(icp, ix, NCAT)
      call eval_probs(ra2, dec2, ix, icat, pmatch, pcoinc)
      net_prob  = max(0.0_double, pmatch - pcoinc)
      xlist%netprob(ix,ncat) = net_prob
      best_prob = max(best_prob, net_prob)
   end do
   if(best_prob > 0.0_double) NMATCHES = NMATCHES + 1
   goodness = goodness + best_prob
end do
! Negate the goodness value as the fitting function MINIMISES the result 
result = -goodness
end subroutine cross_match
!-----------------------------------------------------------------------------
subroutine eval_probs(xra, xdec, ix, icat, pmatch, pcoinc)
real(double), intent(in) :: xra, xdec ! X-ray source RA,DEC [radians]
integer, intent(in)      :: ix        ! No of X-ray object in the list
integer, intent(in)      :: icat      ! No of object in current optical catalogue, NCAT
real(double), intent(out) :: pmatch   ! Probability that X-ray and opt object match
real(double), intent(out) :: pcoinc   ! Probability that this is just a coincidence
!
real(double) :: offset, sigmasq
!
! Compute great-circle distance between X-ray object and optical/IR in radians
!
offset  = gcdist(xra, xdec, CATS(NCAT)%cat_ra(icat), CATS(NCAT)%cat_dec(icat))
sigmasq = xlist%radec_err(ix)**2 + CATS(NCAT)%perr(icat)**2 + SYSERRSQ  ! rads squared
pmatch  = exp(-0.5 * offset**2 / sigmasq)
! To evaluate coincidence need the density of objects in the vicinity
pcoinc = xlist%density(ix,NCAT) * (offset/COINC_RADIUS)**2
end subroutine eval_probs
!-----------------------------------------------------------------------------
subroutine fits_status(subname, status)
! Reports on any non-zero FITS status value with message to standard output
character(*), intent(in) :: subname   ! Name of subroutine or function
integer, intent(in)      :: status    ! FITS status value to check
!
character :: msg*200
!
if(status == 0) RETURN
call ftgerr(status, msg)
write(errStr,'(a,i4,4a)') 'FITSIO error', status, ' in ', trim(subname), &
                     ' - ', trim(msg)
call error('fitsioError')
end subroutine fits_status
!-----------------------------------------------------------------------------
! Locates column number of named column in a FITS table HDU
subroutine get_fits_column(unit, name, ncol, status)
integer, intent(in) :: unit      ! I/O unit of open FITS table
character(*), intent(in) :: name ! Name of the column
integer, intent(out) :: ncol     ! Column number
integer, intent(inout) :: status ! Inherited status
!
if(status /= 0) return
call ftgcno(unit, .false., name, ncol, status)
if(status /= 0) then
   call error('missingColumn', 'Cannot find column '//trim(name) )
end if
end subroutine get_fits_column
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_c(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
character(*), intent(out) :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkys(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_c
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_d(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
real(double), intent(out) :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkyd(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_d
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_e(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
real, intent(out)         :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkye(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_e
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_j(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
integer, intent(out)      :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkyj(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_j
!-----------------------------------------------------------------------------
! Searches header of current HDU for a keyword and returns value
subroutine get_fits_keyword_l(unit, name, value, status)
integer, intent(in)       :: unit   ! I/O unit of open FITS file
character(*), intent(in)  :: name   ! Name of keyword in upper case
logical, intent(out)      :: value  ! Value as string
integer, intent(inout)    :: status ! Inherited status
!
character(50) :: comment ! unused here
if(status /= 0) return
call ftgkyl(unit, name, value, comment, status)
if(status /= 0) then
   call error('missingKeyword', 'FITS keyword missing: '//name)
end if
end subroutine get_fits_keyword_l
!-----------------------------------------------------------------------------
! hist_angle: Finds bad points from histogram of angles between pairs of points/
! Effect is that it resets some values in array xlist%badflag(:)
subroutine hist_angle(xlist, status)
type(XlistType), intent(inout) :: xlist 
integer, intent(inout)         :: status
!
integer, parameter :: nbins = 180, maxpt = 100000
integer :: j, k, n, ibin, nbad, &
   histogram(0:nbins-1), ipoint(0:nbins-1,maxpt), npoints(0:nbins-1), &
   badcount(MAXSRC)
real    :: dx, dy, theta, hmean, hlimit, ibinFloat
!
if(status /= 0) RETURN
histogram = 0
ipoint    = 0
npoints   = 0
!
! First produce a histogram of the angular distibution of position angles
!  between all distinct pairs of points (avoid double-counting pairs).
!  Note that the ipoint array stores the original list indexes of the points
!  forming each pair for each position angle bin for future track-back
!
do j = 1, xlist%nsrc
   do k = j+1, xlist%nsrc
      dy = xlist%dec(j) - xlist%dec(k)
      dx = (xlist%ra(j) - xlist%ra(k)) * cos(xlist%dec(j))
      theta = mod(twopi + atan2(dy,dx), pi) * rtod    ! theta range 0 to 180
      ibinFloat = nbins * theta / 180.0
      if (ibinFloat > nbins) then
        write(errStr, '(a,f11.5,a)') "catcorr_aux: ibin (", ibinFloat &
           , ") exceeds the upper limit.  Contact the code developer."
        call error('ibinTooLarge')
      else if (ibinFloat == nbins) then
        write(errStr, '(a,f11.5,a)') "catcorr_aux: ibin (", ibinFloat &
           , ") is larger than expected."
        call warning('ibinUnexpectedlyLarge')
      end if
      ! ibin = int(nbins * theta / 180.0)
      ibin = min(nbins-1, int(ibinFloat))	! To deal with floating point rounding cases.
      histogram(ibin) = histogram(ibin) + 1
      n = npoints(ibin)
      ipoint(ibin, n+1) = j
      ipoint(ibin, n+2) = k
      npoints(ibin) = n + 2
   end do
end do
!
! Choose suitable threshold for detecting a count excess
!
hmean = real(sum(histogram))/nbins
if(hmean < 20.0) then
   hlimit = poisson_level(hmean, 0.01/nbins)
else  ! Use Gaussian approximation
   hlimit = hmean + dinvnorm(0.01_double/nbins) * sqrt(hmean)
end if
!
! For each position angle at which there is an excess of pairs of points
! mark these by incrementing badcount(:)
!
badcount = 0
do ibin = 0, nbins-1
   if(histogram(ibin) > hlimit) then
! count all bad points
      do j = 1,npoints(ibin)
         k = ipoint(ibin,j)
         badcount(k) = badcount(k) + 1
      end do
   end if
end do
!
! for points with a high enough badpoint count, mark as not good in data
!
nbad = 0
do j = 1,xlist%nsrc
   if(badcount(j) >= 3) then
      xlist%badflag(j) = xlist%badflag(j) + 4
      nbad = nbad + 1
   end if
end do
if(DIAG) write(LOG,*)'hist_angle excluded', nbad, 'more detns, leaving', &
         count(xlist%badflag(1:xlist%nsrc) == 0)
end subroutine hist_angle
!-----------------------------------------------------------------------------
subroutine need_col(unit, colname, coltype, colunits, ncol, status)
! Finds a column by name, else creates new empty one of specified data type
integer, intent(in)      :: unit     ! Unit number of FITS file open to a table
character(*), intent(in) :: colname  ! Name of column
character(*), intent(in) :: coltype  ! Data type, e.g. '1D'
character(*), intent(in) :: colunits ! Physical units string, e.g. "degs"
integer, intent(out)     :: ncol     ! Column number if found/created
integer, intent(inout)   :: status   ! Inherited status
!
integer   :: ncolumns
character :: tunit*8
!
if(status /= 0) RETURN
call ftgcno(unit, .false., colname, ncol, status)
if(status /= 0) then
!
! the column does not exist, so add another column at the end
! 
   status = 0
   call ftgncl(unit, ncolumns, status)
   ncol = ncolumns + 1
   call fticol(unit, ncol, colname, coltype, status)
   write(tunit, '(a,i0)') 'TUNIT', ncol
   call ftukls(unit, tunit, colunits, "physical unit of the column", status)
end if
call fits_status('need_col', status)
end subroutine need_col
!-----------------------------------------------------------------------------
subroutine OBSMLI_read(OBSMLIfile, xlist, status)
! Reads selected columns of an OBSMLI file to the xlist structure.
character(*), intent(in)      :: OBSMLIfile      ! FITS file name of type OBSMLI
type(XlistType), intent(out)  :: xlist           ! selected contents of file
integer, intent(inout)        :: status          ! Inherited status
!
real, parameter :: radec_err_max = 3.0    ! Bad detection if larger (arcsec)
real, parameter :: ep_extent_max = 0.1    ! Bad detection of larger (pixels)
integer :: nrows, nra, ndec, nradec_err, nep_extent, nep_tot, nsrc_num, &
           nlii, nbii, hdutype, iyear, month, date, unit, j, nbad, irow, &
           npn_offax, nm1_offax, nm2_offax
logical :: anyf
character    :: date_obs*20
real(double) :: ra_pnt, dec_pnt
real         :: offax(MAXSRC,3)  ! 3 slices for PN, M1, M2 off-axis angles
!
if(status /= 0) RETURN
xlist%nsrc   = 0
!
! Here open file READONLY - need to reopen later if we wish to update
!
call ftgiou(unit, status)
call ftnopn(unit, OBSMLIfile, 0, status) 
if(status /= 0) &
   call error('missingOBSMLI', 'cannot open file '//trim(OBSMLIfile))
!
! Get following scalars from FITS primary header
!
call get_fits_keyword(unit, 'OBS_ID', xlist%obs_id, status)
call get_fits_keyword(unit, 'RA_PNT', ra_pnt, status)
call get_fits_keyword(unit, 'DEC_PNT', dec_pnt, status)
call get_fits_keyword(unit, 'DATE-OBS', date_obs, status)
! Convert pointing axis coordinates from degrees to radians
xlist%ra_pnt = ra_pnt * dtor
xlist%dec_pnt = dec_pnt * dtor
if(status /= 0) go to 1000
!
! date of observation needed for proper-motion calculation, convert to real year
!
read(date_obs, '(i4,1x,i2,1x,i2)', iostat=status) iyear, month, date
if(status == 0) then
   xlist%year = 2000.0 + (mjd(iyear, month, date) - mjd(2000, 1, 1)) / 365.25
else
   call warning('invalidDateObs', 'error decoding DATE_OBS value')
   if(DIAG) write(LOG,*) 'OBSMLI: error decoding DATE_OBS vale', date_obs, iyear, month, date
   xlist%year = 2000.0
   date_obs = '?'
end if
!
! Move to Header Data Unit 2 and check if these headers exist from some
! earlier run of eposcorr or catcorr.
!
call ftmahd(unit, 2, hdutype, status)
xlist%poscorok  = .false.
xlist%offset    = 0.0
xlist%offerr    = 0.0
xlist%goodness  = 0.0
xlist%nmatches  =   0  
call fits_status('OBSMLI_read/ftgkye', status)
status = 0
!
! Get number of rows, and then column numbers of required columns
!
call ftgnrw(unit, nrows, status)
xlist%nsrc = nrows
call get_fits_column(unit, 'RA',        nra, status)
call get_fits_column(unit, 'DEC',       ndec, status)
call get_fits_column(unit, 'LII',       nlii, status)
call get_fits_column(unit, 'BII',       nbii, status)
call get_fits_column(unit, 'RADEC_ERR', nradec_err, status)
call get_fits_column(unit, 'EP_EXTENT', nep_extent, status)
call get_fits_column(unit, 'EP_TOT',    nep_tot, status)
call get_fits_column(unit, 'SRC_NUM',   nsrc_num, status)
call get_fits_column(unit, 'PN_OFFAX',  npn_offax, status)
call get_fits_column(unit, 'M1_OFFAX',  nm1_offax, status)
call get_fits_column(unit, 'M2_OFFAX',  nm2_offax, status)
!
! Read the arrays in
!
call ftgcvd(unit, nra,        1, 1, nrows, 0.0d0, xlist%ra, anyf, status)
call ftgcvd(unit, ndec,       1, 1, nrows, 0.0d0, xlist%dec, anyf, status)
call ftgcvd(unit, nlii,       1, 1, nrows, 0.0d0, xlist%lii, anyf, status)
call ftgcvd(unit, nbii,       1, 1, nrows, 0.0d0, xlist%bii, anyf, status)
call ftgcve(unit, nradec_err, 1, 1, nrows, 0.0,   xlist%radec_err, anyf, status)
call ftgcve(unit, nep_extent, 1, 1, nrows, 0.0,   xlist%ep_extent, anyf, status)
call ftgcve(unit, nep_tot,    1, 1, nrows, 0.0,   xlist%ep_tot, anyf, status)
call ftgcvj(unit, nsrc_num,   1, 1, nrows, 0,     xlist%src_num, anyf, status)
!
! Read in arrays of off-axis angles, setting null elements to -1.0
!
call ftgcve(unit, npn_offax,  1, 1, nrows, -1.0, offax(:,1), anyf, status)
call ftgcve(unit, nm1_offax,  1, 1, nrows, -1.0, offax(:,2), anyf, status)
call ftgcve(unit, nm2_offax,  1, 1, nrows, -1.0, offax(:,3), anyf, status)
!
! Convert units to radians for RA, DEC, RADEC_ERR
!
xlist%ra(1:nrows)        = xlist%ra(1:nrows)        * dtor
xlist%dec(1:nrows)       = xlist%dec(1:nrows)       * dtor
xlist%lii(1:nrows)       = xlist%lii(1:nrows)       * dtor
xlist%bii(1:nrows)       = xlist%bii(1:nrows)       * dtor
xlist%radec_err(1:nrows) = xlist%radec_err(1:nrows) * ASTOR
! 
! Compute off-axis angle based on RA_PNT, DEC_PNT
!
do irow = 1, nrows
   xlist%offax(irow) = gcdist(xlist%ra(irow), xlist%dec(irow), &
                              xlist%ra_pnt, xlist%dec_pnt)
end do
!
! Get the original position arrays and use as the starting point of the new fit
!
xlist%ra_cc(1:nrows)  = xlist%ra(1:nrows)
xlist%dec_cc(1:nrows) = xlist%dec(1:nrows)
xlist%lii_cc(1:nrows) = xlist%lii(1:nrows)
xlist%bii_cc(1:nrows) = xlist%bii(1:nrows)
call fits_status('OBSMLI_read/ftgcvd', status)
if(status /= 0) go to 1000
!
! Mark rows as bad if RADEC_ERR zero or too large, or source significantly extended.
!
xlist%badflag(1:nrows) = 0
do j = 1,nrows
   if(xlist%ep_extent(j) > ep_extent_max) then
      xlist%badflag(j) = xlist%badflag(j) + 1
   end if
   if(xlist%radec_err(j)*RTOAS > radec_err_max ) then
      xlist%badflag(j) = xlist%badflag(j) + 2
   end if
end do
nbad = count(xlist%badflag(1:nrows) > 0)
1000 continue
call ftclos(unit, status)
call ftfiou(unit, status)
call fits_status('OBSMLI_read', status)
if(DIAG) then
   write(LOG,'(3a,2f10.4,2a,i4,a,i4,a)') 'OBS_ID=', xlist%obs_id, &
    ' target ', xlist%ra_pnt*rtod, xlist%dec_pnt*rtod, ' DATE-OBS=', date_obs, &
    xlist%nsrc, ' x-ray sources but', nbad, ' extd or large perr'               
end if
end subroutine OBSMLI_read
!-----------------------------------------------------------------------------
! OBSMLI_update: writes new positions to FITS file
subroutine OBSMLI_update(obsmlifile, bestcat, astcorr, status)
character(*), intent(in) :: obsmlifile ! to be opened and updated
integer, intent(in)      :: bestcat    ! No of chosen best catalogue
logical, intent(in)      :: astcorr    ! Whether astrometric corrections done
integer, intent(inout)   :: status     ! Inherited status
!
integer :: unit, hdutype, nsrc, ncolra, ncoldec, idate(8), ncolglong, &
           ncolglat, ncolbadflag, ncolnetprob, ncolsyserrcc, &
           nra, ndec, nlii, nbii
character :: datestring*20
real      :: offset, offerr
!
if(status /= 0) return
nsrc = xlist%nsrc
! Open the FITS file for read/write
call ftgiou(unit, status)
call ftnopn(unit, OBSMLIfile, 1, status)
call fits_status('OBSMLI_update/ftopen', status)
!
! Insert history keyword with version number and date/time
!
call date_and_time(values=idate)
write(datestring, "(i4,2('-',i2.2),'T',i2.2,2('-',i2.2))") idate(1:3), idate(5:7)
call ftphis(unit, 'Modified by catcorr v'// version //' at '//datestring, &
            status)
!
! Move to 2nd HDU and check it is the right type i.e. binary table
!
call ftmahd(unit, 2, hdutype, status)
if(hdutype /= 2) then
   call error('obsmliNotTable', 'HDU2 not binary table as expected')
end if
!
! Insert/update header keywords
!
call ftukyd(unit, 'RAOFFSET', xlist%offset(1,bestcat)*RTOAS, 5, &
   'RA correction [arcsec]', status)
call ftukyd(unit, 'DEOFFSET', xlist%offset(2,bestcat)*RTOAS, 5, &
   'DEC correction [arcsec]', status)
call ftukyd(unit, 'ROT_CORR', xlist%offset(3,bestcat)*RTOD, 5, &
   'Rotation correction [degrees]', status)
call ftukyd(unit, 'RAOFFERR', xlist%offerr(1,bestcat)*RTOAS, 5, &
   'Error in RA correction [arcsec]', status)
call ftukyd(unit, 'DEOFFERR', xlist%offerr(2,bestcat)*RTOAS, 5, &
   'Error in Dec correction [arcsec]', status)
call ftukyd(unit, 'ROT_ERR ', xlist%offerr(3,bestcat)*RTOD, 5, &
   'Error in rotation correction [degrees]', status)
offset = sqrt((xlist%offset(1,bestcat)*cos(xlist%dec_pnt))**2 + &
               xlist%offset(2,bestcat)**2) * RTOAS
offerr = sqrt((xlist%offerr(1,bestcat)*cos(xlist%dec_pnt))**2 + &
               xlist%offerr(2,bestcat)**2) * RTOAS
call ftukye(unit, 'POFFSET', offset, 5, 'Field pos correction [arcsec]', status)
call ftukye(unit, 'POFFERR', offerr, 5, 'Field pos error [arcsec]', status)
call ftukye(unit, 'LIK_HOOD', xlist%goodness(bestcat), 3, &
   'Goodness-of-fit statistic', status)
call ftukye(unit, 'LIKHOOD0', xlist%goodzero(bestcat), 3, &
   'Original goodness-of-fit statistic', status)
call ftukyj(unit, 'NMATCHES', xlist%nmatches(bestcat), 'No. of X-ray/ref matches', status)   
call ftukyl(unit, 'POSCOROK', xlist%poscorok, 'Positions updated by catcorr', status)
call ftukyl(unit, 'ASTCORR',  astcorr, 'Positions astrometrically corrected', status)
call ftukys(unit, 'REFCAT',   xlist%catused, 'Catalogue used by catcorr', status)
!
! If astrometric corrections are applied, need to write updated RA, DEC, LII, BII 
! values as well
!
if(astcorr) then
   call get_fits_column(unit, 'RA',  nra, status)
   call get_fits_column(unit, 'DEC', ndec, status)
   call get_fits_column(unit, 'LII', nlii, status)
   call get_fits_column(unit, 'BII', nbii, status)
   call ftpcld(unit, nra,  1, 1, nsrc, xlist%ra(1:nsrc)*rtod, status)
   call ftpcld(unit, ndec, 1, 1, nsrc, xlist%dec(1:nsrc)*rtod, status)
   call ftpcld(unit, nlii, 1, 1, nsrc, xlist%lii(1:nsrc)*rtod, status)
   call ftpcld(unit, nbii, 1, 1, nsrc, xlist%bii(1:nsrc)*rtod, status)
end if
!
! write updated positions converted to degrees from RA_CC/DEC_CC values
!
call need_col(unit, 'RA_CORR', '1D', 'deg', ncolra, status)
call ftpcld(unit, ncolra,  1, 1, nsrc, xlist%RA_CC(1:nsrc)*rtod, status)
!
call need_col(unit, 'DEC_CORR', '1D', 'deg', ncoldec, status)
call ftpcld(unit, ncoldec, 1, 1, nsrc, xlist%DEC_CC(1:nsrc)*rtod, status)
! Ditto for galactics
call need_col(unit, 'LII_CORR', '1D', 'deg', ncolglong, status)
call ftpcld(unit, ncolglong, 1, 1, nsrc, xlist%lii_cc(1:nsrc)*rtod, status)
call need_col(unit, 'BII_CORR', '1D', 'deg', ncolglat, status)
call ftpcld(unit, ncolglat,  1, 1, nsrc, xlist%bii_cc(1:nsrc)*rtod, status)
!
! Add new column SYSERRCC [arcsec] for individual Catcor posn errors
!
call need_col(unit, 'SYSERRCC', '1E', 'arcsec', ncolsyserrcc, status)
call ftpcle(unit, ncolsyserrcc, 1, 1, nsrc, xlist%syserrcc(1:nsrc), status)
!
! In diagnostic mode, write two new columns to the file
!
if(DIAG) then
   call need_col(unit, 'BADFLAG', '1J', ' ', ncolbadflag, status)
   call ftpclj(unit, ncolbadflag, 1, 1, nsrc, xlist%badflag(1:nsrc), status)
   call need_col(unit, 'NETPROB', '1E', ' ', ncolnetprob, status)
   call ftpcle(unit, ncolnetprob, 1, 1, nsrc, xlist%netprob(1:nsrc,NCAT), &
               status)
end if
!
! Delete existing header keyword SYSERRCC if exists as now duplicates column
!
call ftdkey(unit, 'SYSERRCC', status)
status = 0           ! ignore failure of last call

call ftclos(unit, status)
call ftfiou(unit, status)
if(DIAG) then
  write(LOG,*) 'OBSMLI_update: correction', xlist%offset(1:2,bestcat)*RTOAS, &
     xlist%offset(3,bestcat)*RTOD
end if
call fits_status('OBSMLI_update', status)
end subroutine OBSMLI_update
!-----------------------------------------------------------------------------
! print_detlist: in diagnostic mode prints detection details to log file
subroutine print_detlist(icat)
integer, intent(in) :: icat  ! Catalogue number
!
integer      :: ix, icp, iopt
real(double) :: ra2, dec2, pmatch, pcoinc, net_prob, offset0, offset1
!
NCAT = icat
write(log,'(10a)') &
 ' ix     raunc    decunc    racorr   deccorr perr cp  iopt', &
 ' pmatch pcoinc pnet rauncof decuncof offax syserrcc cat=', cats(icat)%extname
do ix = 1, xlist%nsrc
   if(xlist%ncp(ix,icat) == 0 .or. xlist%badflag(ix) > 0) CYCLE
   call rot_shift(xlist%ra_pnt, xlist%dec_pnt, xlist%offset(:,icat), &
                  xlist%ra(ix), xlist%dec(ix), ra2, dec2)
   do icp = 1, xlist%ncp(ix,icat)
      iopt = xlist%listcp(icp, ix, icat)
      call eval_probs(ra2, dec2, ix, iopt, pmatch, pcoinc)
      net_prob  = max(0.0_double, pmatch-pcoinc)
      if(net_prob >= 1e-5) then
         offset0 = gcdist(xlist%ra(ix), xlist%dec(ix), CATS(icat)%cat_ra(iopt), &
                         CATS(icat)%cat_dec(iopt))
         offset1 = gcdist(ra2, dec2, cats(ncat)%cat_ra(iopt), cats(ncat)%cat_dec(iopt))
         write(log, '(i3,4f10.5, f5.2,i3,i6, 3f7.4, 4f6.2)') &
            ix, xlist%ra(ix)*rtod, xlist%dec(ix)*rtod, ra2*rtod, dec2*rtod, &
            xlist%radec_err(ix)*rtoas, xlist%ncp(ix,icat), iopt, &
            pmatch, pcoinc, net_prob, &
            (xlist%ra(ix)-cats(icat)%cat_ra(iopt))*cos(dec2)*rtoas, &
            (xlist%dec(ix)-cats(icat)%cat_dec(iopt))*rtoas, &
            xlist%offax(ix)*RTOD*60.0, xlist%syserrcc(ix)
      end if
   end do
end do
end subroutine print_detlist
!-----------------------------------------------------------------------------
subroutine print_summary
integer :: icat
write(log, '(10a)') 'refcat   nobjs nmatches goodzero goodness  raoff decoff  rotn', &
                  ' raofer decofer roterr'
do icat = 1,MAXCAT
   write(log, '(a,i6,i9, 2f9.2, 6f7.2   )') &
      cats(icat)%extname, cats(icat)%nobj, xlist%nmatches(icat), &
      xlist%goodzero(icat), xlist%goodness(icat), &
      xlist%offset(1:2,icat)*rtoas, xlist%offset(3,icat)*rtod, &
      xlist%offerr(1:2,icat)*rtoas, xlist%offerr(3,icat)*rtod
end do
end subroutine print_summary
!-----------------------------------------------------------------------------
! rotvec: rotates a 3-vector to a new frame
subroutine rotvec(vin, ctos, vout)
real(double), intent(in)  :: vin(3)     ! Input 3-vector
real(double), intent(in)  :: ctos(3,3)  ! Rotation matrix
real(double), intent(out) :: vout(3)    ! Output 3-vector
!
integer :: j
!
do j = 1,3
   vout(j) = sum(vin(:) * ctos(:,j))
end do
end subroutine rotvec
!--
end module catcorr_aux
